# from . import *
from .tc import TestCaseBase
from .devices import DutBase, DeviceBase

from typing import *
import json
from pathlib import Path
from PyQt5.QtCore import QThread

from pyqt_templates import *
from object_info import ObjectInfo
from private_values import PrivateJson

from importlib import import_module

import TESTCASES
import DEVICES


# =====================================================================================================================
class Exx__TcItemNotFound(Exception):
    pass


class Exx__TcItemType(Exception):
    pass


class Exx__TcSettingsIncorrect(Exception):
    pass


# =====================================================================================================================
class TpMultyDutBase(QThread):
    signal__tp_finished = pyqtSignal()
    signal__tp_stop = pyqtSignal()

    # SETTINGS ------------------------------------
    # DIRPATH_TPS: Union[str, Path] = "TESTPLANS"
    DIRPATH_TCS: Union[str, Path] = "TESTCASES"
    # DIRPATH_DEVS: Union[str, Path] = "DEVICES"

    TCS: Dict[Union[str, Type[TestCaseBase]], Optional[bool]]    # settings
    # {
    #     Tc1: True,
    #     Tc2: True
    # }

    DUT_CLS: Type[DutBase]

    # DEVICES: List[Union[str, Type[DeviceBase]]]    # settings
    # [
    #     Dev1,
    #     Dev2
    # ]

    # AUXILIARY -----------------------------------
    DUTS: List[DutBase]    # autogenerated
    tc_active: Optional[Type[TestCaseBase]] = None

    def __init__(self):
        super().__init__()
        # self.DIRPATH_TPS: Path = Path(self.DIRPATH_TPS)
        self.DIRPATH_TCS: Path = Path(self.DIRPATH_TCS)
        # self.DIRPATH_DEVS: Path = Path(self.DIRPATH_DEVS)

        self.reinit()

    def reinit(self, tcs: Optional[Dict[Type[TestCaseBase], Optional[bool]]] = None) -> Optional[NoReturn]:
        # TCS --------------------------------------------------------------
        tcs = tcs or dict(self.TCS)
        self.TCS = {}
        for item, using in tcs.items():
            # print(dir(TESTCASES))
            if isinstance(item, str):   # filename
                # tc_cls = import_module(item, "TESTCASES").TestCase    # not working!
                # tc_cls = getattr(TESTCASES, item).TestCase      # not working
                tc_cls = import_module(f"TESTCASES.{item}").TestCase
                if not tc_cls:
                    msg = f"[ERROR] file not found[{item=}] in TESTCASES/"
                    raise Exx__TcItemNotFound(msg)
                tc_cls.NAME = item
            elif isinstance(type(item), type) and issubclass(item, TestCaseBase):
                tc_cls = item
                # msg = f"[ERROR] DONT USE IT!"
                # raise Exception(msg)
            else:
                msg = f"[ERROR] type is inconvenient [{item=}]"
                raise Exx__TcItemType(msg)

            tc_cls.SKIP = not using
            self.TCS.update({tc_cls: using})

            settings_filepath = self.DIRPATH_TCS.joinpath(f"{item}.json")
            if settings_filepath.exists():
                # print(f"{settings_filepath=} EXISTS")
                tc_cls.SETTINGS = PrivateJson(_filepath=settings_filepath)
            else:
                # print(f"{settings_filepath=} NOT_EXISTS")
                pass

        # print(f"{tc_cls.SETTINGS=}")

        # DUTS --------------------------------------------------------------
        self.DUTS = []
        self.duts_generate()
        self._duts_mark_presented()
        self._duts_results_tp_init()

    # TCS -----------------------------------------------------------
    @property
    def tcs_active(self) -> List[TestCaseBase]:
        result = []

        if self.tc_active:
            for dut in self.DUTS:
                tc_dut = dut.TP_RESULTS[self.tc_active]
                if not tc_dut.skip_tc_dut:
                    result.append(tc_dut)
        return result

    # DUTS -----------------------------------------------------------
    def duts_generate(self) -> None:
        # raise NotImplemented
        pass

    def _duts_mark_presented(self) -> None:
        for dut in self.DUTS:
            dut.mark_present()

    def _duts_results_tp_init(self) -> None:
        for dut in self.DUTS:
            dut.TP_RESULTS = dict()
            for tc in self.TCS:
                dut.TP_RESULTS.update({tc: tc(dut)})

    def _duts_results_tc_clear(self) -> None:
        for dut in self.DUTS:
            dut.results_tc_clear()

    # --------------------------------------------------------------------
    def terminate(self) -> None:
        super().terminate()

        # TERMINATE CHILDS!!! ---------------------
        # ObjectInfo(self.currentThread()).print()    # cant find childs!!!

        # finish current ----------------------------
        if self.tc_active:
            self.tc_active.terminate_all()
            self.tc_active = None

        self.signal__tp_finished.emit()

    # RUN ----------------------------------------------------------------
    def run(self) -> None:
        for tc in self.TCS:
            self.tc_active = tc
            tc.run_all(self.DUTS)

        # FINISH TPlan ---------------------------------------------------
        self.tc_active = None
        self.signal__tp_finished.emit()


# =====================================================================================================================
